{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [],
   "source": [
    "class VMCustomized(object):\n",
    "    def __init__(self, register):\n",
    "        self.r1_values = []\n",
    "        self.ticks = 0\n",
    "        self.ip = -1\n",
    "        self.ip_binding = None\n",
    "        self.register = register\n",
    "        self.instructions = []\n",
    "        self.opcodes = {\n",
    "            \"addr\": self.addr,\n",
    "            \"addi\": self.addi,\n",
    "            \"mulr\": self.mulr,\n",
    "            \"muli\": self.muli,\n",
    "            \"banr\": self.banr,\n",
    "            \"bani\": self.bani,\n",
    "            \"borr\": self.borr,\n",
    "            \"bori\": self.bori,\n",
    "            \"setr\": self.setr,\n",
    "            \"seti\": self.seti,\n",
    "            \"gtir\": self.gtir,\n",
    "            \"gtri\": self.gtri,\n",
    "            \"gtrr\": self.gtrr,\n",
    "            \"eqir\": self.eqir,\n",
    "            \"eqri\": self.eqri,\n",
    "            \"eqrr\": self.eqrr\n",
    "        }\n",
    "\n",
    "    def load(self, instructions):\n",
    "        self.instructions = []\n",
    "        for instruction in instructions:\n",
    "            if \"#ip\" == instruction[:3]:\n",
    "                _, ip_str = instruction.split(\" \")\n",
    "                self.ip_binding = int(ip_str)\n",
    "            else:\n",
    "                op, a, b, c = instruction.split(\" \")\n",
    "                self.instructions.append([op, int(a), int(b), int(c)])\n",
    "\n",
    "    def get_debug_state(self):\n",
    "        return \" \".join([\n",
    "            f\"ip={self.ip}\",\n",
    "            str(self.register),\n",
    "            \" \".join(str(i) for i in self.instructions[self.ip])\n",
    "        ])\n",
    "        \n",
    "    def serialize_state(self):\n",
    "        return \"-\".join(str(r) for r in self.register)\n",
    "        \n",
    "    def run(self, debug=False):\n",
    "        while True:\n",
    "            try:\n",
    "                self.ip += 1\n",
    "                \n",
    "                if self.ip == 28:                    \n",
    "                    print(self.register[1])                    \n",
    "                    if (self.register[1] in self.r1_values):\n",
    "                        break\n",
    "                    self.r1_values.append(self.register[1])                    \n",
    "                \n",
    "                debug_state = self.get_debug_state()\n",
    "                self.step()\n",
    "                self.ticks += 1\n",
    "                if debug:\n",
    "                    print(debug_state, str(self.register))\n",
    "\n",
    "            except:\n",
    "                break\n",
    "                \n",
    "        return 0\n",
    "        \n",
    "    def step(self):       \n",
    "        if self.ip >= len(self.instructions):\n",
    "            return False\n",
    "                \n",
    "        op, a, b, c = self.instructions[self.ip]\n",
    "        self.execute(op, a, b, c)\n",
    "            \n",
    "        return True\n",
    "        \n",
    "    def execute(self, opcode, a, b, c):\n",
    "        if self.ip_binding != None:\n",
    "            self.register[self.ip_binding] = self.ip\n",
    "            \n",
    "        decode_int = {\n",
    "            0 : self.eqir,\n",
    "            1 : self.seti,\n",
    "            2 : self.eqri,\n",
    "            3 : self.eqrr,\n",
    "            4 : self.addi,\n",
    "            5 : self.setr,\n",
    "            6 : self.gtrr,\n",
    "            7 : self.gtri,\n",
    "            8 : self.muli,\n",
    "            9 : self.bori,\n",
    "            10: self.bani,\n",
    "            11: self.borr,\n",
    "            12: self.gtir,\n",
    "            13: self.banr,\n",
    "            14: self.addr,\n",
    "            15: self.mulr\n",
    "        }\n",
    "        \n",
    "        decode_str = {\n",
    "            \"eqir\" : self.eqir,\n",
    "            \"seti\" : self.seti,\n",
    "            \"eqri\" : self.eqri,\n",
    "            \"eqrr\" : self.eqrr,\n",
    "            \"addi\" : self.addi,\n",
    "            \"setr\" : self.setr,\n",
    "            \"gtrr\" : self.gtrr,\n",
    "            \"gtri\" : self.gtri,\n",
    "            \"muli\" : self.muli,\n",
    "            \"bori\" : self.bori,\n",
    "            \"bani\" : self.bani,\n",
    "            \"borr\" : self.borr,\n",
    "            \"gtir\" : self.gtir,\n",
    "            \"banr\" : self.banr,\n",
    "            \"addr\" : self.addr,\n",
    "            \"mulr\" : self.mulr\n",
    "        }\n",
    "        \n",
    "        if (type(opcode) is str):\n",
    "            assert(decode_str[opcode](a,b,c))\n",
    "        elif (type(opcode) is int):\n",
    "            assert(decode_int[opcode](a,b,c))\n",
    "        else:\n",
    "            raise Exception(f\"opcode {opcode} is neither int nor str\")\n",
    "        \n",
    "        if self.ip_binding != None:\n",
    "            self.ip = self.register[self.ip_binding]\n",
    "        \n",
    "    def any_invalid_registers(self, regs):\n",
    "        result = True\n",
    "\n",
    "        for reg in regs:\n",
    "            result = result & (reg < len(self.register))\n",
    "\n",
    "        return not result\n",
    "\n",
    "    def test_opcode(self, opcode_name, initial_register_state, a, b, c, final_register_state):\n",
    "        self.register = [ reg for reg in initial_register_state ]\n",
    "        result = self.opcodes[opcode_name](a, b, c)\n",
    "        return result and (self.register == final_register_state)\n",
    "\n",
    "    # addr (add register) stores into register C the result of adding register A and register B.\n",
    "    def addr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a,b,c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] + self.register[b]\n",
    "        return True\n",
    "\n",
    "    # addi (add immediate) stores into register C the result of adding register A and value B.\n",
    "    def addi(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False    \n",
    "\n",
    "        self.register[c] = self.register[a] + b\n",
    "        return True\n",
    "\n",
    "    # mulr (multiply register) stores into register C the result of multiplying register A and register B.\n",
    "    def mulr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] * self.register[b]\n",
    "        return True\n",
    "\n",
    "    # muli (multiply immediate) stores into register C the result of multiplying register A and value B.\n",
    "    def muli(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] * b\n",
    "        return True\n",
    "\n",
    "    # banr (bitwise AND register) stores into register C the result of the bitwise AND of register A and register B.\n",
    "    def banr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] & self.register[b]\n",
    "        return True\n",
    "\n",
    "    # bani (bitwise AND immediate) stores into register C the result of the bitwise AND of register A and value B.\n",
    "    def bani(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] & b\n",
    "        return True\n",
    "\n",
    "    # borr (bitwise OR register) stores into register C the result of the bitwise OR of register A and register B.\n",
    "    def borr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] | self.register[b]\n",
    "        return True\n",
    "\n",
    "\n",
    "    # bori (bitwise OR immediate) stores into register C the result of the bitwise OR of register A and value B.\n",
    "    def bori(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] | b\n",
    "        return True\n",
    "\n",
    "    # setr (set register) copies the contents of register A into register C. (Input B is ignored.)\n",
    "    def setr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a]\n",
    "        return True\n",
    "\n",
    "    # seti (set immediate) stores value A into register C. (Input B is ignored.)\n",
    "    def seti(self, a, b, c):\n",
    "        if self.any_invalid_registers([c]):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = a\n",
    "        return True\n",
    "\n",
    "    # gtir (greater-than immediate/register) sets register C to 1 if value A is greater than register B. Otherwise, register C is set to 0.\n",
    "    def gtir(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, b)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(a > self.register[b])\n",
    "        return True\n",
    "\n",
    "    # gtri (greater-than register/immediate) sets register C to 1 if register A is greater than value B. Otherwise, register C is set to 0.\n",
    "    def gtri(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, a)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] > b)\n",
    "        return True\n",
    "\n",
    "    # gtrr (greater-than register/register) sets register C to 1 if register A is greater than register B. Otherwise, register C is set to 0.\n",
    "    def gtrr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] > self.register[b])\n",
    "        return True\n",
    "\n",
    "    # eqir (equal immediate/register) sets register C to 1 if value A is equal to register B. Otherwise, register C is set to 0.\n",
    "    def eqir(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, b)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(a == self.register[b])\n",
    "        return True\n",
    "\n",
    "    # eqri (equal register/immediate) sets register C to 1 if register A is equal to value B. Otherwise, register C is set to 0.\n",
    "    def eqri(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, a)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] == b)\n",
    "        return True\n",
    "\n",
    "    # eqrr (equal register/register) sets register C to 1 if register A is equal to register B. Otherwise, register C is set to 0.\n",
    "    def eqrr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a,b,c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] == self.register[b])\n",
    "        return True\n",
    "    \n",
    "    def disassemble(self):\n",
    "        return \"\\n\".join(\n",
    "            f\"{i:02}: \" + self.disassemble_instr(instr)\n",
    "            for i, instr in enumerate(self.instructions)\n",
    "        )\n",
    "    \n",
    "    def disassemble_instr(self, instr):\n",
    "        op, a_, b_, c = instr\n",
    "        \n",
    "        a = self.disassemble_reg(a_, True)\n",
    "        ra = self.disassemble_reg(a_, False)\n",
    "        b = self.disassemble_reg(b_, True)\n",
    "        rb = self.disassemble_reg(b_, False)\n",
    "        rc = self.disassemble_reg(c, False)\n",
    "        \n",
    "        diss_instr = {\n",
    "            \"addr\": \"{ra} + {rb}\",\n",
    "            \"addi\": \"{ra} + {b}\",\n",
    "            \"mulr\": \"{ra} * {rb}\",\n",
    "            \"muli\": \"{ra} * {b}\",\n",
    "            \"banr\": \"{ra} & {rb}\",\n",
    "            \"bani\": \"{ra} & {b}\",\n",
    "            \"borr\": \"{ra} | {rb}\",\n",
    "            \"bori\": \"{ra} | {b}\",\n",
    "            \"setr\": \"{ra}\",\n",
    "            \"seti\": \"{a}\",\n",
    "            \"gtrr\": \"{ra} > {rb} ? 1 : 0\",\n",
    "            \"gtri\": \"{ra} > {b} ? 1 : 0\",\n",
    "            \"gtir\": \"{a} > {rb} ? 1 : 0\",\n",
    "            \"eqir\": \"{a} == {rb} ? 1 : 0\",\n",
    "            \"eqri\": \"{ra} == {b} ? 1 : 0\",    \n",
    "            \"eqrr\": \"{ra} == {rb} ? 1 : 0\",    \n",
    "        }\n",
    "        \n",
    "        return f\"{rc} = \" + diss_instr[op].format(a=a, ra=ra, b=b, rb=rb)\n",
    "    \n",
    "    def disassemble_reg(self, reg, immediate):\n",
    "        if self.ip_binding != None and reg == self.ip_binding:\n",
    "            return \"IP\"\n",
    "        elif immediate:\n",
    "            return str(reg)\n",
    "        return f\"R{reg}\""
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {
    "scrolled": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "00: R1 = 123\n",
      "01: R1 = R1 & 456\n",
      "02: R1 = R1 == 72 ? 1 : 0\n",
      "03: IP = R1 + IP\n",
      "04: IP = 0\n",
      "05: R1 = 0\n",
      "06: R2 = R1 | 65536\n",
      "07: R1 = 8725355\n",
      "08: R5 = R2 & 255\n",
      "09: R1 = R1 + R5\n",
      "10: R1 = R1 & 16777215\n",
      "11: R1 = R1 * 65899\n",
      "12: R1 = R1 & 16777215\n",
      "13: R5 = 256 > R2 ? 1 : 0\n",
      "14: IP = R5 + IP\n",
      "15: IP = IP + 1\n",
      "16: IP = 27\n",
      "17: R5 = 0\n",
      "18: R3 = R5 + 1\n",
      "19: R3 = R3 * 256\n",
      "20: R3 = R3 > R2 ? 1 : 0\n",
      "21: IP = R3 + IP\n",
      "22: IP = IP + 1\n",
      "23: IP = 25\n",
      "24: R5 = R5 + 1\n",
      "25: IP = 17\n",
      "26: R2 = R5\n",
      "27: IP = 7\n",
      "28: R5 = R1 == R0 ? 1 : 0\n",
      "29: IP = R5 + IP\n",
      "30: IP = 5\n"
     ]
    }
   ],
   "source": [
    "puzzle_input_str = open(\"puzzle_input/day21.txt\").read()\n",
    "\n",
    "def disassemble():\n",
    "    instructions = puzzle_input_str.split(\"\\n\")\n",
    "    vm = VMCustomized([0] * 6)\n",
    "    vm.load(instructions)\n",
    "    print(vm.disassemble())\n",
    "\n",
    "# run_program(puzzle_input_str, [0] * 6)\n",
    "disassemble()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "instructions = puzzle_input_str.split(\"\\n\")\n",
    "vm = VMCustomized(registers)\n",
    "vm.load(instructions)\n",
    "vm.run(debug)"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
